# comic-get.pl - Utility for fetching comics from the internet
# Copyright (C) 2004-2006 Tommi Saviranta <wnd@iki.fi>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

package ComicGet::Repository;

=head1 NAME
ComicGet::Repository - object to access repository-related stuff
=cut

use strict;
use warnings FATAL => 'all';

use ComicGet::Config;
use ComicGet::Packages;
use ComicGet::Fetch;
use ComicGet::Database;
use ComicGet::Index;
use ComicGet::Files;

use vars qw($SEARCH $SHOW);

$SEARCH = 0;
$SHOW = 1;

my $file_no = 0;


=item print_stats
Print network statistics.
=cut
sub print_stats() {
	my ($time, $recv_bytes, $recv_files) = &ComicGet::Web::stats();
	$time = 1 if ($time < 1);
	my $bytes = &ComicGet::Common::human_readable($recv_bytes);
	my $recv_rate = &ComicGet::Common::human_readable($recv_bytes / $time);
	printf("Fetched %s in %.1f s (%s/s, %s files)\n",
			$bytes,
			$time,
			$recv_rate,
			$recv_files);
}


=item install_single
Installs a single package (comic strip).

0: Package
1: Language (optional)
r: Installed language (undef if no _new_ package was intalled)
=cut
sub install_single($) {
	my ($pack, $lang) = @_;

	my @repos;
	foreach my $file (&ComicGet::Config::get_packages_files()) {
		my ($entry, $iter, $repo_id);

		($repo_id, $iter) = &ComicGet::Packages::read($file);
		if (! defined($repo_id)) {
			if (-e $file) {
				print "E: \"$file\" broken!\n";
				return undef;
			}
			print "E: Cannot read \"$file\": " .
				"$ComicGet::Common::error\n";
			return undef;
		}
		while (1) {
			($entry, $iter) =
				&ComicGet::Packages::next_entry($iter);
			last if (! defined($entry));
			next if (! &ComicGet::Packages::check_entry_val(
					$entry, "Package", $pack, 1));

			my $wlang;
			if (! defined($lang)) {
				$wlang = &ComicGet::Packages::get_entry_field(
						$entry, "Language");
			} else {
				$wlang = $lang;
			}
			if (&ComicGet::Packages::check_entry_val(
					$entry, "Language", $wlang)) {
				push(@repos, "$repo_id\t$wlang");
			}
		}
	}

	if ($#repos == -1) {
		print "E: Couldn't find package $pack";
		print " for language $lang" if (defined($lang));
		print "\n";
		return undef;
	}

	foreach my $line (@repos) {
		my ($id, $lang) = split(/\t/, $line);
		my $old_repo = &ComicGet::Database::package_repo($pack, $lang);
		if (defined($old_repo)) {
			print "I: $pack ($lang) already installed " .
				"($old_repo)\n";
			return undef;
		}
	}

	if ($#repos != 0) {
		print "More than one candidate for $pack";
		print " ($lang)" if (defined($lang));
		print ":\n";
		for (my $i = 0; $i <= $#repos; $i++) {
			my ($repo_id, $tlang) = split(/\t/, $repos[$i]);
			my $repo_uri = &ComicGet::Index::source_by_id($repo_id);
			die("Repo URI missing") if (! defined($repo_uri));
			print "  $i: $tlang through $repo_uri\n";
		}

		my $n;
		do {
			print "Which one should I install? " .
				"Press enter to cancel.\n";
			$n = &ComicGet::Common::ask();
		} while (length($n) > 0 && ($n =~ /[^\d]/ ||
				$n < 0 || $n > $#repos));
		if ($n eq "") {
			printf("Skipping installation of %s%s...\n",
					$pack,
					defined($lang) ? " ($lang)" : "");
			return undef;
		}
		$repos[0] = $repos[$n];
	}

	my @info = split(/\t/, $repos[0]);
	my $repo_uri = &ComicGet::Index::source_by_id($info[0]);
	$lang = $info[1];
	my $uri = &ComicGet::Files::rule_uri($repo_uri, $pack, $lang);
	my $rulefile = &ComicGet::Files::rule($pack, $lang);
	# TODO check http_get response
	my $ruledata = &ComicGet::Web::http_get($uri, 80);

	my $target = &ComicGet::Fetch::get_file($ruledata);
	die("Rulefile has no filename") if (! defined($target));
	my @file_info = &ComicGet::Database::file_installed($target);
	if (defined($file_info[0])) {
		print "File $target already installed by " .
			"$file_info[0] ($file_info[1])\n";
		while (1) {
			print <<EOF;
If you would like to install the package anyway enter new filename here.
If you want to cancel, just press enter.
EOF
			$target = &ComicGet::Common::ask();
			last if (length($target) == 0);

			@file_info = &ComicGet::Database::file_installed(
					$target);
			last unless (defined($file_info[0]));
			printf("%s installed by %s (%s)\n",
					$target,
					$file_info[0],
					$file_info[1]);
		}
		if (length($target) == 0) {
			print "Skipping $pack ($lang)...\n";
			return undef;
		}
	}

	&ComicGet::Database::insert($info[0], $pack, $info[1], $target);

	&ComicGet::Common::save_file($rulefile, $ruledata)
			or die("Cannot save \"$rulefile\": " .
				&ComicGet::Common::error);

	return $info[1];
}


=item install
Install package(s).

@: List of packages
=cut
sub install(@) {
	foreach my $pack (@_) {
		my ($ppack, $lang) = split(/=/, $pack);
		$lang = &install_single($ppack, $lang);
		next if (! defined($lang));
		
		foreach my $line (&ComicGet::Database::get_database()) {
			my %info = &ComicGet::Database::line_to_hash($line);
			if ($info{'name'} eq $ppack && $info{'lang'} eq $lang) {
				&upgrade_single(\%info);
				last;
			}
		}
	}

	&print_stats();
}


=item remove_single
Remove single package.

0: Package(=language)
r: 1 if success
=cut
sub remove_single($) {
	my $pack_raw = shift;
	my @files;
	my ($pack, $lang) = split(/=/, $pack_raw);

	my @inst = &ComicGet::Database::get_installed($pack, $lang);
	if ($#inst == -1) {
		if (defined($lang)) {
			print "E: $pack ($lang) not installed.\n";
		} else {
			print "E: $pack not installed.\n";
		}
		return 0;
	}

	if ($#inst > 0) {
		print "More than one package matching $pack installed:\n";
		for (my $i = 0; $i <= $#inst; $i++) {
			my %pack = %{($inst[$i])};
			my $lang = $pack{'lang'};
			my $repo_uri = &ComicGet::Index::source_by_id(
					$pack{'repo'});
			printf "  $i: $lang through $repo_uri\n";
		}

		my $n;
		do {
			print "Which one should I remove? " .
				"Empty line cancels.\n";
			$n = &ComicGet::Common::ask();
		} while (length($n) > 0 && ($n =~ /[^\d]/ ||
				$n < 0 || $n > $#inst));

		if ($n eq "") {
			print "Skipping removal of $pack...\n";
			return 0;
		}
		$inst[0] = $inst[$n];
	}
	
	my %t = %{($inst[0])};
	$lang = $t{'lang'};
	my $rulefile = &ComicGet::Files::rule($pack, $lang);

	if (! -e $rulefile) {
		my $t = unlink("$rulefile");
		print "\nW: Cannot remove \"$rulefile\": $!\n" if ($t == 0);
	} else {
		print "\nW: Rulefile \"$rulefile\" missing. Ignoring...\n";
	}

	print "$pack ($lang) removed.\n";

	&ComicGet::Database::remove($pack, $lang);
	return 1;
}


=item remove
Remove one or more packages.

@: Packages
=cut
sub remove(@) {
	foreach my $pack (@_) {
		&remove_single($pack);
	}
}


=item is_installed
See if package is already installed.

0: package
1: language (optional)
r: 1 if installed, 0 if not
=cut
sub is_installed($$) {
	my $package = shift;
	my $lang = shift;
	$lang = "_[a-z][a-z]_[A-Z][A-Z]" if (! defined($lang));
	die("remove me");

	my @files = &ComicGet::Common::read_dir("$ComicGet::Config::home",
			"^$package$lang.*\.rule\$");
	return defined($files[0]) ? 1 : 0;
}


=item find_source
Find source.

0: Package file
r: Source
=cut
sub find_source($) {
	&ComicGet::Config::read();
	my ($file) = $_[0] =~ /(.*)_Packages/;
	foreach my $source (@ComicGet::Config::sources) {
		my $cfile = &ComicGet::Common::uri_to_file($source);
		return $source if ($cfile eq $file);
	}
	return undef;
}


=item get_inst_candidate

0: package
r: repository
=cut
sub get_inst_candidate($) {
}


=item upgrade_single
Upgrade single package.

0: Package info
=cut
sub upgrade_single(\%) {
	my %info = %{(shift)};
	my $file = &ComicGet::Files::rule($info{'name'}, $info{'lang'});
	my $rules = &ComicGet::Common::read_file($file);
	(defined($rules)) || die("Cannot read \"$file\": " .
		$ComicGet::Common::error);

	&ComicGet::Fetch::follow_rules($rules, $info{'file'});
}


=item upgrade
Upgrade (refresh) packages.
=cut
sub upgrade() {
	foreach my $line (&ComicGet::Database::get_database()) {
		my %pack = &ComicGet::Database::line_to_hash($line);
		&upgrade_single(\%pack);
	}

	&print_stats();
}


=item import_repository
Import repository for use.

0: Repository URI
=cut
sub import_repository($) {
	my $repo_uri = shift;

	$file_no++;
	print "Get: $file_no $repo_uri Packages\n";

	# TODO check http_get response
	my $uri = &ComicGet::Files::packages_uri($repo_uri);
	my $data = &ComicGet::Web::http_get($uri);

	my ($repo_id, $unused) = &ComicGet::Packages::get_repo_id($data);

	my $file = &ComicGet::Files::md5sums($repo_id);
	# Touch file
	open(FILE, "> $file") or die("Cannot write \"$file\": $!");
	close(FILE);

	&ComicGet::Index::insert($repo_id, $repo_uri);
}


=item update
Update rules.
=cut
sub update() {
	&ComicGet::Config::read();

	# Check index
	foreach my $src (@ComicGet::Config::sources) {
		my $id = &ComicGet::Index::id_by_source($src);
		next if (defined($id));
		&import_repository($src);
	}

	foreach my $repo_uri (@ComicGet::Config::sources) {
		my $repo_id = &ComicGet::Index::id_by_source($repo_uri);
		my $lf_mod = 0;

		my $lf_md5 = &ComicGet::Files::md5sums($repo_id);
		my $lp_md5 = &ComicGet::Common::read_file($lf_md5);
		if (! defined($lp_md5) && -e $lf_md5) {
			print "E: Cannot read \"$lp_md5\": " .
				"&ComicGet::Common::error\n";
			next;
		}
		my %l_md5 = &page_to_hash($lp_md5) if (defined($lp_md5));

		$file_no++;
		print "Get: $file_no $repo_uri md5sums\n";

		my $uri = &ComicGet::Files::md5sums_uri($repo_uri);
		my $rp_md5 = &ComicGet::Web::http_get("$uri", 80);
		my %r_md5 = &page_to_hash($rp_md5) if (defined($rp_md5));

		if (&ComicGet::Common::safe_compare($l_md5{"Packages"},
				$r_md5{"Packages"}) == 0) {
			&update_packages($repo_id);
			$l_md5{"Packages"} = $r_md5{"Packages"};
			$lf_mod = 1;
		}
		if (&ComicGet::Common::safe_compare($l_md5{"rules.md5sums"},
					$r_md5{"rules.md5sums"}) == 0) {
			&update_rules($repo_id);
			$l_md5{"rules.md5sums"} = $r_md5{"rules.md5sums"};
			$lf_mod = 1;
		}
		if ($lf_mod == 1) {
			&ComicGet::Common::save_file($lf_md5, $rp_md5)
					or die("Cannot save \"$lf_md5\": " .
						&ComicGet::Common::error);
		}
	}

	&print_stats();
}


=item save_hash
Save hash.

0: File
1: Hash
r: 1 on success
=cut
sub save_hash($%) {
	my $file = shift;
	my %hash = @_;
	my $data = "";
	while (my ($rule, $md5) = each %hash) {
		$data .= "$md5  $rule\n" if (defined($md5));
	}
	return &ComicGet::Common::save_file($file, $data);
}


=item update_packages
Update package list from certain repository.

0: Repository identifier
=cut
sub update_packages($) {
	my $repo_id = shift;
	my $repo_uri = &ComicGet::Index::source_by_id($repo_id);

	$file_no++;
	print "Get: $file_no $repo_uri Packages\n";

	my $uri = &ComicGet::Files::packages_uri($repo_uri);
	my $file = &ComicGet::Files::packages($repo_id);

	my $data = &ComicGet::Web::http_get($uri);

	&ComicGet::Common::save_file($file, $data)
			or die("Cannot save \"$file\": " .
				&ComicGet::Common::error);
}


=item update_rules
Update package rules from certain repository.

0: Repository identifier
=cut
sub update_rules($) {
	my $repo_id = shift;
	my $repo_uri = &ComicGet::Index::source_by_id($repo_id);
	my ($uri, $file, $data);
	my (%l_md5, %r_md5);

	$file_no++;
	print "Get: $file_no $repo_uri rules.md5sums\n";

	$uri = &ComicGet::Files::rules_md5sums_uri($repo_uri);
	$file = &ComicGet::Files::rules_md5sums($repo_id);

	$data = &ComicGet::Common::read_file($file);
	%l_md5 = &page_to_hash($data);
	
	$data = &ComicGet::Web::http_get($uri);
	%r_md5 = &page_to_hash($data);

	while (my ($rule, $md5) = each %l_md5) {
		next if (&ComicGet::Common::safe_compare(
				$md5, $r_md5{$rule}) == 1);
		next if (! defined($r_md5{$rule}));

		my @files = &ComicGet::Common::read_dir( # TODO
				"$ComicGet::Config::home", "^$rule\$");
		next if ($#files == -1);
	
		$file_no++;
		print "Get: $file_no $repo_uri $rule\n";

		my $data = &ComicGet::Web::http_get("$repo_uri/$rule"); # TODO
		my $rfile = "$ComicGet::Config::home/$rule"; # TODO
		&ComicGet::Common::save_file($rfile, $data)
				or die("Cannot write \"$rfile\": " .
					$ComicGet::Common::error);
		$l_md5{$rule} = $r_md5{$rule};
	}

	&ComicGet::Common::save_file($file, $data)
			or die("Cannot write \"$file\": " .
				$ComicGet::Common::error);
}


=item page_to_hash
Convert packages file into hash.

0: Page content.
1: Hash of md5sums
=cut
sub page_to_hash($) {
	my %hash;
	return %hash if (! defined($_[0]));
	foreach $_ (split(/[\n\r]/, $_[0])) {
		my @t = split(/\s+/);
		(defined($t[0]) && defined($t[1])) or die("Bad file!");
		$hash{"$t[1]"} = $t[0];
	}
	return %hash;
}


=item is_searched
See if search candidate matches given rules.

0: Data
1: Mode ($SEARCH/$SHOW)
@: Rules

r: 1 if match
=cut
sub is_searched($$@) {
	my $data = shift;
	my $mode = shift;
	my @rules = @_;
	my $match;

	if ($mode == $SEARCH) {
		foreach my $rule (@rules) {
			return 0 if (! ($data =~ /$rule/));
		}
		return 1;
	} elsif ($mode == $SHOW) {
		foreach my $rule (@rules) {
			return 1 if ($data =~ /Package: $rule\n/);
		}
		return 0;
	} else {
		die("Invalid mode: $mode");
	}
}


=item print_entry
Print entry information.

0: Entry
1: Mode
=cut
sub print_entry($$) {
	my ($data, $mode) = @_;

	if ($mode == $SEARCH) {
		my ($pack) = $data =~ /Package: (.*?)\n/;
		my ($short) = $data =~ /Description: (.*?)\n/;
		my ($lang) = $data =~ /Language: (.*?)\n/;
		print "$pack ($lang) - $short\n";
	} elsif ($mode == $SHOW) {
		$data =~ s/\n .\n/\n\n/sg;
		print "$data\n";
	}
}


=item search
Search packages.

0: mode: ($SEARCH/$SHOW)
@: search strings (AND)
=cut
sub search($@) {
	my $mode = shift;
	my @rules = @_;

	&ComicGet::Config::read();
	foreach my $file (&ComicGet::Config::get_packages_files()) {
		my ($id, $iter) = &ComicGet::Packages::read($file);
		my $entry;
		if (! defined($id)) {
			if (-e $file) {
				print "E: File \"$file\" broken!\n";
				print "Bailing out...\n";
				exit(1);
			}
			print "E: Cannot read \"$file\": " .
				"$ComicGet::Common::error\n";
			next;
		}
		while (1) {
			($entry, $iter) =
				&ComicGet::Packages::next_entry($iter);
			last if (! defined($entry));
			
			if (&is_searched($entry, $mode, @rules)) {
				&print_entry($entry, $mode);
				print "\n" if ($mode == $SHOW);
			}
		}
	}
}


1;
